// FILE: cush.c
// A command shell with minimal functionality.
// WRITTEN BY: Grant Macklem (grant.macklem@colorado.edu)
// DATE LAST MODIFIED: February 1, 2001

#include <stdlib.h>    // provides EXIT_SUCCESS, free, malloc
#include <unistd.h>    // provides fork, execvp, pipe
#include <stdio.h>     // provides printf
#include <string.h>    // provides strcasecmp
#include <sys/types.h> // Needed for wait
#include <sys/wait.h>  // Provides wait
#include <ctype.h>     // Provides isspace
#include <signal.h>    // Provides kill
#include <sys/stat.h>  // Needed for open
#include <fcntl.h>     // Needed for open

#define FALSE 0
#define TRUE 1
#define MAX_ARGS 22    // Set this to 2 higher than the max you want
#define MAX_LEN 800
//#define DEBUG

// Prototypes
void free_memory(char **argv);
int parse_input(char buffer[], char** argv);
char* eat_white(char buffer[]);
int check_redirection(char *argv[]);
int check_pipe(char *argv[]);
int setup_output(char filename[]);
int setup_input(char filename[]);
int setup_append(char filename[]);
int check_in_out(int setup(char*), char** argv, char *ptr);


int main(int argc, char** argv)
{
    char *argvc[MAX_ARGS];       // Argv for the child process
    char buffer[MAX_LEN+1];      // Reading buffer
    char *ampersand; // Will point to the ampersand if it exists
    int bg;          // Whether the process should be run in the background
    pid_t pid;       // Process ID of the child
    
    while(TRUE)
    {
	printf("cush> ");
	fgets(buffer, MAX_LEN, stdin);

	// Check for backgrounding
	ampersand = strchr(buffer, '&');
	bg = (ampersand != NULL) ? TRUE : FALSE;
	if (bg)
	    ampersand[0] = '\n';   // Remove ampersand
	
	if (strcasecmp(buffer, "exit\n") == 0)
	    break;   // Exit the shell (user typed exit)
	
	if (buffer[0] == '\n')
	    continue; // Do nothing, reprint prompt (user pressed enter)

	// Parse input
	if (parse_input(buffer, argvc) != 0)
	{ // If an error, print message, free memory, and loop
	    printf("Too many arguments. Limit is %d.\n", MAX_ARGS - 2);
	    free_memory(argvc);
	    continue;
	}

	// Check for the kill command
	if (strcasecmp(argvc[0], "kill") == 0)
	    if(kill((pid_t)atoi(&argvc[1][1]), atoi(argvc[2])) != 0)
	    { // If an error, print message, free memory, and loop
		perror("Error while killing");
		free_memory(argvc);
		continue;
	    }

#ifdef DEBUG
#else
	pid = fork();
	if (pid < 0)
	    perror("Unable to create child process");
	else if (pid == 0)  // in child process
	{
#endif
	    // Check for piping
	    if (check_pipe(argvc) == -1)  // Too many pipes or error
		exit(-1);
	    // No pipes, otherwise check_pipe wouldn't have returned
            // Check for errors in redirection.
	    if (check_redirection(argvc) < 0)
		exit(-1);
	    if (execvp(&argvc[0][0], argvc) == -1)  // Execute the command
	    {
		perror(argvc[0]);
		exit(-1);
	    }
#ifdef DEBUG
#else
	}
	else  // in parent process
#endif
	    if (bg == FALSE)
		while (pid != wait(NULL)); // Wait for completion
	    else
		printf("[%d] %s\n", (int)pid, argvc[0]);
	
	free_memory(argvc);
    }

    return EXIT_SUCCESS;
}


// This function will free the memory allocated during parsing of the
// input string.  If a pipe is found, a pointer in the middle is changed
// to NULL, so a simple checking for NULL won't work.  When originally
// parsing, the final two pointers are set to null.  If only one is set
// to null when freeing, there must be more afterwards.
void free_memory(char **argv)
{
    int i = -1;           // Loop variable to clear memory
    while (argv[++i])     // Until the last void pointer
	free(argv[i]);    // Clear memory allocated for parameters
    if (argv[i+1] != NULL)  // If next pointer is not null, a pipe was used
	while (argv[++i])   // Free memory until the next void pointer
	    free(argv[i]); 
}


int parse_input(char buffer[], char** argv)
{
    char *pos = buffer;      // Pointer into buffer
    char *space;       //  = strchr(buffer, ' ');
    size_t size;       // Size of argument to malloc
    int i = 0;         // Current argument parsed
    int brk = FALSE;   // Whether to break out of the loop
    int ret_code = 0;  // Return code
    
    pos = eat_white(pos);  // eliminate white space before first command
    while(!brk)
    {
	space = strchr(pos, ' ');

	if (space == NULL)
	{
	    // Could be at the end of the line
	    if (pos[0] == '\0')
		break;
	    space = strchr(pos, '\n');

	    // Could be only one word on the command line
	    if (space == NULL)
		space = strchr(pos, '\0');
	    
	    brk = TRUE;
	}

	// Create space for the argument plus null termination
	size = space - pos;
	argv[i] = (char*) malloc(size + 1);
	strncpy(&argv[i][0], pos, size);  // sizeof(char) = 1, no mult
	argv[i][size] = '\0';
	pos = space;      // Set the buffer past the current argument
	pos = eat_white(pos);
	if (++i > MAX_ARGS - 2)
	{
	    --i;
	    ret_code = 1;
	    break;
	}
    }
    
    // Set the 2 last pointers to null.  If a pipe is used, only one
    // pointer will be NULL and we can free the other memory.
    argv[i] = NULL;
    argv[i+1] = NULL;
    return ret_code;
}


// This function will eat all the white space between two parameters
// on the command line.  It will return a pointer to the next
// nonwhitespace character in the buffer.
char* eat_white(char buffer[])
{
    int i = -1;
    while isspace((int)buffer[++i]);
    return &buffer[i];
}


// Thus function will check for any redirection token in the argument
// strings and set up the process' input and/or output as specified.
int check_redirection(char *argv[])
{
    int i, j, out_found = 0, in_found = 0;
    char *ptr;   // Pointer into the argument at the redirection token.

    // No increment, see note at bottom of loop
    for (i = 0; argv[i] != NULL;)
    {
        // ************************************************************
        // Check for output redirection
        // ************************************************************
	ptr = strchr(argv[i], '>');   // Find the output redirection token
	if ( (ptr != NULL) && (ptr[1] != '>') ) // if it's not appending
	{
	    // Check to see if previous output redirection set up
	    if (++out_found > 1)
	    {
		fprintf(stderr, "Output may only be directed into one file.\n");
		return -1;
	    }
	    else
	    {
		if (check_in_out(setup_output, &argv[i], ptr) != 0)
		    return -1;
	    }
	    continue;
	}
	
	
        // ************************************************************
        // Check for input redirection
        // ************************************************************
	ptr = strchr(argv[i], '<');   // Find the output redirection token
	if (ptr != NULL)
	{
	    // Check to see if previous input redirection set up
	    if (++in_found > 1)
	    {
		fprintf(stderr, "Input may only be directed from one file.\n");
		return -1;
	    }
	    else
	    {
		if (check_in_out(setup_input, &argv[i], ptr) != 0)
		    return -1;
	    }
	    continue;
	}

	
        // ************************************************************
        // Check for appending output redirection
        // ************************************************************
	ptr = strchr(argv[i], '>');   // Find the output redirection token
	if ((ptr != NULL) && (ptr[1] == '>')) // Make sure it's appending
	{
	    // Check to see if previous output redirection set up
	    if (++out_found > 1)
	    {
		fprintf(stderr, "Output may only be directed into one file.\n");
		return -1;
	    }
	    else
	    {
		if (strlen(argv[i]) == 2)
		{ // Token was the entire argument ">>"
		    if (setup_append(argv[i+1]) != 0)
		    {
			return -1;
		    }		    		    
			
		    // Fix the argv array to eliminate the redirection tokens
		    j = i;
		    free(argv[j]);     // The ">>"
		    free(argv[j+1]);   // The filename
		    do
		    {
			argv[j] = argv[j+2];
		    } while (argv[j++]);
		}
		else if (ptr[2] == '\0')
		{ // Token occured at the end of an argument "grep>>"
		    ptr[0] = '\0';   // Remove token
		    if (setup_append(argv[i+1]) != 0)
		    {
			return -1;
		    }		    		    
			
		    // Fix the argv array to eliminate the redirection tokens
		    j = i+1;
		    free(argv[j]);   // The filename
		    do
		    {
			argv[j] = argv[j+1];
		    } while (argv[j++]);
		}
		else if (ptr == argv[i])
		{ // Token occured at the beginning ">>out.txt"		    
		    if (setup_append(&argv[i][2]) != 0)
		    {
			return -1;
		    }		    		    
			
		    // Fix the argv array to eliminate the redirection tokens
		    j = i;
		    free(argv[j]);    // The filename
		    do
		    {
			argv[j] = argv[j+1];
		    } while (argv[j++]);
		}
		else
		{ // Token occurred in the middle "grep>>out.txt"
		    ptr[0] = '\0';    // Terminate string at token
		    if (setup_append(&ptr[2]) != 0)
		    {
			return -1;
		    }		    		    
		}
	    }
	    continue;
	}

	// Only increment i if nothing was parsed.  If a token was parsed
	// it is effectively removed from the command string and the
	// current point must be reevaluated.
	++i;
    }
    
    // Everything went okay
    return 0;
}


// This function checks for a pipe in the arguments.  If one is found,
// a pipe is created between the two commands and they are executed
// with their respective arguments.  If a pipe is found, this function
// will not return unless there is an error.
int check_pipe(char *argv[])
{
    int i, j, pipe_count = 0;  // count of pipes
    char *ptr;   // Pointer into the argument at the pipe token.
    char *tmp, *tmp2, *tmp3; // Temporary pointers for reorganizing list
    int arg2;    // index to argument lists for second command
    int pid, pfd[2];   // second process id, pipe file descriptors

    // No increment, see note at bottom of loop
    for (i = 0; argv[i] != NULL;)
    {
        // ************************************************************
        // Check for piping
        // ************************************************************
	ptr = strchr(argv[i], '|');   // Find pipe token
	if (ptr != NULL)
	{
	    if (++pipe_count > 1)
	    {
		fprintf(stderr, "You may only pipe once.\n");
		return -1;
	    }
	    
	    if (strlen(argv[i]) == 1)
	    { // Token was the entire argument "|"
		arg2 = i+1;   // Set second command argument list
		// pointer to the argument after the pipe.
		free(argv[i]);      // Free location of pipe
		argv[i] = NULL;     // and set it to null.
	    }
	    else if (ptr[1] == '\0')
	    { // Token occured at the end of an argument "grep|"
		ptr[0] = '\0';   // Remove token
		// Fix the argv array to insert a NULL pointer
		j = i+1;
		tmp = argv[j];
		argv[j++] = NULL;  // End command 1's argument list
		arg2 = j;    // Set second command argument list
		// Insert null termination of arg list into array
		do
		{
		    tmp2 = argv[j];
		    argv[j] = tmp;
		    tmp = tmp2;
		} while (argv[j++]);
	    }
	    else if (ptr == argv[i])
	    { // Token occured at the beginning "|less"
		// Copy string to new location, omitting pipe
		tmp = (char*)malloc(strlen(argv[i]));
		strcpy(tmp, &argv[i][1]);
		// End command 1's argument list
		free(argv[i]);
		argv[i] = NULL;

		j = i+1;
		arg2 = j;    // Set second commands argument list
		// Insert null termination of arg list into array
		do
		{
		    tmp2 = argv[j];
		    argv[j] = tmp;
		    tmp = tmp2;
		} while (argv[j++]);
	    }
	    else
	    { // Token occurred in the middle "grep|less"
		ptr[0] = '\0';    // Terminate string at token
		tmp = (char*)malloc(strlen(&ptr[1]));
		strcpy(tmp, &ptr[1]);
		tmp2 = argv[i+1];
		argv[i+1] = NULL;
		j = i+2;
		arg2 = j;      // Set second command's argument list
		// Insert null termination of arg list into array
		do
		{
		    tmp3 = argv[j];
		    argv[j] = tmp;
		    tmp = tmp2;
		    tmp2 = tmp3;
		} while (argv[j++]);
	    }    
	}
	i++;
    }

    if (pipe_count == 0)  // no pipes found
	return 0;

    // Set up the pipe
    if (pipe(pfd) == -1)
    {
	perror("pipe");
	return -1;
    }

    pid = fork();
    switch(pid)
    {
    case -1:
	perror("Unable to create child process");
	return -1;
	break;
    case 0:  // In child process
        // Close unwanted end of the pipe
	close(pfd[0]);
	// Close standard out
	close(1);
	// Duplicate output port
	dup(pfd[1]);
	close(pfd[1]);
	if (execvp(&argv[0][0], argv) == -1)  // Execute the command
	{
	    perror(argv[0]);
	    return -1;
	}
	break;
    default:  // In parent process
	// Close unwanted end of the pipe
	close(pfd[1]);
	// Close standard in
	close(0);
	// Duplicate input port
	dup(pfd[0]);
	close(pfd[0]);
	if (execvp(&argv[arg2][0], &argv[arg2]) == -1)  // Execute the command
	{
	    perror(argv[arg2]);
	    return -1;
	}
	break;
    }
    return 0;
}



// This function sets up stdout for the child process to write output
// to a file.
int setup_output(char filename[])
{
    int fid;

    // Open file write only and create it if it doesnt exist.  Set
    // permissions 664
    fid = open(filename, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR \
	       | S_IRGRP | S_IWGRP | S_IROTH);
    if (fid == -1)
    {
	perror("An error occured while redirecting output");
	return -1;
    }
    else
    {
	close(1);   // Close standard out
	dup(fid);   // Duplicate the open file, putting it in stdout
	close(fid); // Close the original file
    }
    return 0;
}


// This function sets up stdin for the child process to input from a
// file.
int setup_input(char filename[])
{
    int fid;

    // Open the file for read only.
    fid = open(filename, O_RDONLY);

    if (fid == -1)
    {
	perror("An error occured while redirecting input");
	return -1;
    }
    else
    {
	close(0);     // Close standard in
	dup(fid);     // Duplicate the open file, putting it in stdin
	close(fid);   // Close the original file
    }
    return 0;
}


// This function sets up stdout for the child process to append output
// to a file.
int setup_append(char filename[])
{
    int fid;

    // Open file write only and create it if it doesnt exist.  Move the
    // pointer to the end of file and set permissions 664.
    fid = open(filename, O_WRONLY | O_APPEND | O_CREAT, S_IRUSR | \
	       S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
    if (fid == -1)
    {
	perror("An error occured while redirecting output");
	return -1;
    }
    else
    {
	close(1);      // Close stdout
	dup(fid);      // Duplicate the open file, puttint it in stdout
	close(fid);    // Close the original file
    }
    return 0;
}


// This function will check for both input and output redirection
// calling the appropriate setup function for the specific type
// of redirection as specified.
int check_in_out(int setup(char*), char** argv, char *ptr)
{
    int j;
    
    if (strlen(argv[0]) == 1)
    { // Token was the entire argument ">"
	if (setup(argv[1]) != 0)
	{
	    return -1;
	}		    		    
        // Fix the argv array to eliminate the redirection tokens
	j = 0;
	free(argv[j]);     // The ">"
	free(argv[j+1]);   // The filename
	do
	{
	    argv[j] = argv[j+2];
	} while (argv[j++]);
    }
    else if (ptr[1] == '\0')
    { // Token occured at the end of an argument "grep>"
	ptr[0] = '\0';   // Remove token
	if (setup(argv[1]) != 0)
	{
	    return -1;
	}		    		    
	// Fix the argv array to eliminate the redirection tokens
	j = 1;
	free(argv[j]);   // The filename
	do
	{
	    argv[j] = argv[j+1];
	} while (argv[j++]);
    }
    else if (ptr == argv[0])
    { // Token occured at the beginning ">out.txt"
	if (setup(&argv[0][1]) != 0)
	{
	    return -1;
	}		    		    
	// Fix the argv array to eliminate the redirection tokens
	j = 0;
	free(argv[j]);    // The filename
	do
	{
	    argv[j] = argv[j+1];
	} while (argv[j++]);
    }
    else
    { // Token occurred in the middle "grep>out.txt"
	ptr[0] = '\0';    // Terminate string at token
	if (setup(&ptr[1]) != 0)
	{
	    return -1;
	}		    		    
    }    
    return 0;
}
